#include "StdAfx.h"
#include "ecran.h"
#include "Tortue.h"
#include <vector>
#include <list>
#include <map>
using namespace std;

typedef void (Tortue:: *FnMembreTortue) ( TypeParametre);

class Command{
public:
	virtual void execute( Tortue &  t ) = 0;
};

class CompositeCommand :  public Command{
public:
		void push_back( Command * c ){ commandes.push_back(c ); }
		void execute( Tortue &  t )
		{
			for (int i = 0; i< commandes.size() ; i++){
				commandes[i]->execute(t);
			}
		}
private :
	vector<Command* > commandes;
};

class BasicCommand : public Command{
public:
	BasicCommand( FnMembreTortue f , double v ) : m_f(f), m_value(v) {}
	void execute( Tortue &  t )
	{
		// utiliser le ptr sur fonction pour la tortue
		(t.*m_f)(m_value);
	}
private :
	// la fonction  appeler
	FnMembreTortue m_f;
	double m_value;
};

class RepeatCommand : public Command{
public:
	RepeatCommand( Command * pc, int nbRepeats) 
		: m_pc(pc), 
		m_nbRepeats(nbRepeats) {}
	void execute( Tortue &  t )
	{
		for (int i = 0; i < m_nbRepeats; i++)
			m_pc->execute(t);
	}
private :
	// la Commande  appeler
	Command * m_pc;
	int m_nbRepeats;
};

const TypeParametre rayon = 100.0;
const int nbPas= 100;
const double PI = 3.1415;
const int nbArcs = 3;

void dessine(HWND hWnd)
{
	Ecran e(hWnd);

	Tortue tortue (&e, e.getXCentre() , e.getYCentre() );

	BasicCommand avance(&Tortue::avance,	2.0*PI*rayon/(nbArcs*nbPas) );
	BasicCommand tourneADroite (&Tortue::tourne,	-2.0*PI/(nbArcs*nbPas ) );
	
	CompositeCommand unSegmentADroite;
	unSegmentADroite.push_back(&avance);
	unSegmentADroite.push_back(&tourneADroite );

		
	RepeatCommand unArcADroite(&unSegmentADroite, nbPas);

	CompositeCommand unSegmentAGauche;
	unSegmentAGauche.push_back(&avance);
	BasicCommand tourneAGauche (&Tortue::tourne,	2.0*PI/(nbArcs*nbPas ) );
	unSegmentAGauche.push_back(&tourneAGauche);
	RepeatCommand unArcAGauche(&unSegmentAGauche, nbPas);

	CompositeCommand uneRosace;
	uneRosace.push_back(&unArcADroite);
	BasicCommand tourne120(&Tortue::tourne, 2.0*PI/3.0 );
	uneRosace.push_back ( & tourne120);
	uneRosace.push_back(&unArcAGauche);
	uneRosace.push_back ( & tourne120);

	RepeatCommand laFigure( &uneRosace, 3);
	laFigure.execute(tortue);

}
