// Purpose.  dynamic_cast<T> lab
//
// Motivation.  "My Component classes do not know that Composites exist.
// They provide no help for navigating Composites, nor any help for
// altering the contents of a Composite.  This is because I would like the
// base class (and all its derivatives) to be reusable in contexts that do
// not require Composites.  When given a base class pointer, if I
// absolutely need to know whether or not it is a Composite, I will use
// dynamic_cast() to figure this out.  In those cases where dynamic_cast()
// is too expensive, I will use a Visitor." [Robert Martin]
//
// Assignment.
// o Replace the old-style C cast in main() with dynamic_cast<T>
// o While you are at it, add a test for each dynamic_cast<T> to ensure that
//   the returned pointer is not null
// o Add some extra code to main() that assigns a Primitive* to a Component*,
//   uses dynamic_cast<T> on the Component*, and produces the second set of
//   output
// o To eliminate the compiler warning, go to Project - Settings - C/C++ tab -
//   ProjectOptions, and add /GR at the end

#include <iostream>
#include <vector>
using namespace std;

class Component {
   int value;
public:
   Component( int val )    { value = val; }
   virtual void traverse() { cout << value << " "; }
};

class Primitive : public Component { public:
   Primitive( int val ) : Component( val ) { }
};

class Composite : public Component {
   vector<Component*> children;
public:
   Composite( int val ) : Component( val ) { }
   void add( Component* ele ) { children.push_back( ele ); }
   /*virtual*/ void traverse() {
      Component::traverse();
      for (int i=0; i < children.size(); i++)
         children[i]->traverse();
}  };

void main( void ) {
   Component*  nodes[3];
   // The type of Composite* is "lost" when the object is assigned to a
   // Component*
   nodes[0] = new Composite(1);
   nodes[1] = new Composite(2);
   nodes[2] = new Composite(3);

   ((Composite*)nodes[0])->add( nodes[1] ); 
   ((Composite*)nodes[0])->add( nodes[2] );
   ((Composite*)nodes[0])->add( new Primitive(4) );
   ((Composite*)nodes[1])->add( new Primitive(5) );
   ((Composite*)nodes[1])->add( new Primitive(6) );
   ((Composite*)nodes[2])->add( new Primitive(7) );
        
   for (int i=0; i < 3; i++) {
      nodes[i]->traverse();
      cout << endl;
}  }

// 1 2 5 6 3 7 4 
// 2 5 6 
// 3 7 

// 7 is NOT a Composite*
// 1 2 5 6 3 7 4 
// 2 5 6 
// 3 7 
