// Purpose.  RTTI and "reflection" lab answer

#include <iostream>
#include <vector>
#include <string>
using namespace std;

class Command { public: virtual void execute() = 0; };

class Cloneable {
   static vector<Cloneable*> prototypes;
public:
   static void addPrototype( Cloneable* obj ) {
      prototypes.push_back( obj );
   }
   static Cloneable* findAndClone( string name ) {
      for (int i=0; i < prototypes.size(); i++)
         if (prototypes[i]->getName() == name)
            return prototypes[i]->clone();
      cout << name << " not found\n";
      return 0;
   }
   static void report() {
      cout << "Classes registered: ";
      for (int i=0; i < prototypes.size(); i++)
         cout << prototypes[i]->getName() << ' ';
      cout << '\n';
   }

   string getName() {                               // Template Method
      string str = getTypeidName();
      // convert the string "class ABC *" to "ABC"
      return str.substr( 6, str.size() - 8 );
   }
   virtual ~Cloneable()               { }
   virtual string     getTypeidName() = 0;          // placeholder
   virtual Cloneable* clone()         = 0;
};

vector<Cloneable*> Cloneable::prototypes;

class This : public Cloneable, public Command { public:
   /*virtual*/ ~This()                    { cout << "dtor-This "; }
   /*virtual*/ string     getTypeidName() { return typeid( this ).name(); }
   /*virtual*/ Cloneable* clone()         { return new This(); }
   /*virtual*/ void       execute()       { cout << "This: execute\n"; }
};
class That : public Cloneable, public Command { public:
   /*virtual*/ ~That()                    { cout << "dtor-That "; }
   /*virtual*/ string     getTypeidName() { return typeid( this ).name(); }
   /*virtual*/ Cloneable* clone()         { return new That(); }
   /*virtual*/ void       execute()       { cout << "That: execute\n"; }
};
class TheOther : public Cloneable, public Command { public:
   /*virtual*/ ~TheOther()                { cout << "dtor-TheOther "; }
   /*virtual*/ string     getTypeidName() { return typeid( this ).name(); }
   /*virtual*/ Cloneable* clone()         { return new TheOther(); }
   /*virtual*/ void       execute()       { cout << "TheOther: execute\n"; }
};

void initializePrototypes() {
   Cloneable::addPrototype( new This() );
   Cloneable::addPrototype( new That() );
   Cloneable::addPrototype( new TheOther() );
}

void main( int argc, char* argv[] ) {
   vector<Cloneable*> objects;
   Cloneable* obj;
   Command*   cmd;

   initializePrototypes();
   Cloneable::report();

   for (int i=1; i < argc; i++)
      if (obj = Cloneable::findAndClone( argv[i] ))
         objects.push_back( obj );

   for (i=0; i < objects.size(); i++)
      if (cmd = dynamic_cast<Command*>( objects[i] ))
         cmd->execute();

   for (i=0; i < objects.size(); i++) delete objects[i];
}

// D:\projects\AdvCpp\debug> advcpp Garbage This That Nothing This TheOther
// Classes registered: This That TheOther
// Garbage not found
// Nothing not found
// This: execute
// That: execute
// This: execute
// TheOther: execute
// dtor-This dtor-That dtor-This dtor-TheOther
