// Purpose.  Template and operator-> lab
//
// Discussion.  It is a common challenge to need to wrap function invocations
// within pairs of prefix and suffix functionality -
//               grab-lock   do-something   release-lock
//       begin-transaction   do-something   end-transaction
//             trace-entry   do-something   trace-exit
// "Overloading operator-> has long been a popular way of specifying a prefix."
// "One nonintrusive way of doing something 'later' is to create an object of
// a class that has a destructor to do that something." [Stroustrup, C++ Report,
// June 2000, pp9-16]  The solution below is "intrusive".  Classes X and Y have
// to be modified to make it work.  This lab demonstrates Stroustrup's simple-
// minded approach.  His complete approach is considerably more involved.
//
// Assignment.
// o Create a template class that holds a pointer to some TBD type
// o Add a constructor to initialize the private pointer attribute
// o Define a class internal to your template class
// o The "inner" class needs the same attribute and constructor as the "outer"
//   class
// o Additionally, its destructor should call suffix()
// o The inner class' operator-> should return the private pointer attribute
// o Add an operator-> to your outer class that: calls prefix(), and returns
//   (by value) an instance of the inner class
// o Modify main() to create the X and Y objects on the heap, and wrap them in
//   instances of your template class
// o Remove the calls to prefix() and suffix() from f() and g()

// le problme du sandwich (ou du dcorateur)
// ajouter un pr-traitement et un post traitement  une fonction
// le moyen le plus courant est la surcharge de l'oprateur -> ( la smartpointeur)
// cration de la classe Wrappeur "BeforeAfter" (ptr + Ctor + Dtor)
// ajout de la classe interne "After" et des constituants : ptr, Ctor, Dtor, ->()
// l'oprateur -> de la classe Wrappeur doit appeler le pr-traitement 
//	 et rendre possible le post-traitement
//   la solution est de passer "par valeur" un After qui appelle le traitement et est ensuite dtruit
//   la destruction de l'objet after dclenche le post-traitement
// modifier la "main" pour construire un wrappeur sur X et un sur Y
// Attention  bien faire l'allocation par "new"
// Appeler les fonctions que l'on pourra dbarasser des prefix/suffix
#include <iostream>
using namespace std;


class X {
public:
	bool f() { /*prefix();*/  cout << "X::f()";  /*suffix();*/  return true; }
};

class Y {
public:
   void g() { /*prefix();*/  cout << "Y::g()";  /*suffix();*/ }
};


void main( void ) {
/*   X object1;
   if (object1.f()) cout << "f() is true\n";
   Y object2;
   object2.g();
*/
	BeforeAfter<X> wrap1(new X);
	   if (wrap1->f()) cout << "f() is true\n";
	BeforeAfter<Y> wrap2(new Y);
	wrap2->g();
}

// prefix X::f() suffix
// f() is true
// prefix Y::g() suffix
