#define SOLUCE
// Purpose.  priority_queue lab
//
// Discussion.  "A priority_queue is a queue in which each element is given a
// priority that controls the order in which the elements get to be top().  By
// default, the priority_queue simply compares elements using the '<' operator
// and pop() returns the largest element." [Stroustrup, p478]

// Etudier, Compiler et lancer l'exemple suivant
// les Commandes sont munies d'un nombre alatoire reprsentant la priorit
// un ensemble de push et de pop montre que l'extraction se fait d'abord sur les lments
// suprieurs : rem il faut avoir un oprateur '<' pour cela
// le rsultat est toutefois incomplet...
// ... voir la suite de l'exercice plus bas

#ifndef SOLUCE
#include <iostream>
#include <queue>        // priority_queue
#include <ctime>
using namespace std;

struct Command {
   int priority;
   Command( int pri = 0 ) { priority = pri; }
   friend bool operator< ( const Command& lhs, const Command& rhs ) {
      return lhs.priority < rhs.priority; }
   friend ostream& operator<< ( ostream& os, const Command& obj ) {
      return os << obj.priority; }
};

void main( void ) {
   priority_queue<Command> que;
   srand( time(0) );

   cout << "pushing:  ";
   for (int i=0, num; i < 6; i++) {
      cout << (num = rand() % 12) << "  ";
      que.push( Command( num ) );
   } cout << endl;

   cout << "popping:  ";
   for (i=0; i < 3; i++, que.pop()) cout << que.top() << "  ";
   cout << endl;

   cout << "pushing:  ";
   for (i=0, num; i < 6; i++) {
      cout << (num = rand() % 12) << "  ";
      que.push( Command( num ) );
   } cout << endl;

   cout << "popping:  ";
   for (i=0; i < 3; i++, que.pop()) cout << que.top() << "  ";
   cout << endl;

   cout << "pushing:  ";
   for (i=0, num; i < 6; i++) {
      cout << (num = rand() % 12) << "  ";
      que.push( Command( num ) );
   } cout << endl;

   cout << "popping:  ";
   for ( ; que.size(); que.pop()) cout << que.top() << "  ";
   cout << endl;
}

// pushing:  9  11  10  11  0  7
// popping:  11  11  10             // 9 7 0
// pushing:  8  4  10  4  2  10
// popping:  10  10  9              // 8 7 4 4 2 0
// pushing:  1  8  8  11  10  6
// popping:  11  10  8  8  8  7  6  4  4  2  1  0


#else 

// Discussion.  The example above does not demonstrate whether the order of in-
// sertion is preserved.  The example below shows that it is NOT.  Intuition
// might suggest that the output should be - 1-8 3-8 5-8 7-8 9-8 2-5 4-5 6-5
// 8-5 10-5.  "The order in which elements with equal priority come to the head
// of the queue is not defined." [Stroustrup, p479]
//
// Assignment.
// o Write your own priority queue class that preserves order of insertion
// o Decide on a standard library container class to use (the school solution
//   uses a list<T>)
// o Make your class a template class
// o You might consider changing the behavior of pop() so that it returns what
//   top() returns

// La classe Command est modifie pour se souvenir des ordres d'insertion en plus de la priorit
// On remarque que l'ordre n'est pas conserv (7-8 avant 5-8)
//Le but de l'exercice est de crer une priority_queue 
//	qui conserve l'ordre des insertions  priorit gale
// on utilisera une liste car des insertions sont  prvoir au milieu
// ajouter donc l'include correspondant
// crer une classe template PriorityQueue possdant un membre list<T>
// implmenter les fonctions pop et push
// tout est dans le push qui scanne la liste pour trouver le point d'insertion 
//	avant d'effectuer l'ajout
// On pourra liminer la mthode "top" et retourner le rsultat par "pop"


#include <iostream>
#include <queue>        // priority_queue
#include <list>
using namespace std;

class Command {
public:
   Command() { id = next++;   priority = id % 2 ? 8 : 5; }
   friend bool operator< ( const Command& left, const Command& right ) {
      return left.priority < right.priority; }
   friend ostream& operator<< ( ostream& os, const Command& cmd ) {
      return os << cmd.id << "-" << cmd.priority; }
private:
   int id, priority;
   static int next;
};
int Command::next = 1;


void main( void ) {
   PriorityQueue<Command> pq;
//   priority_queue<Command> pq;
   for (int i=0; i < 10; i++)
      pq.push( Command() );
   while ( ! pq.empty()) {
      cout << pq.pop() << " ";
      // cout << pq.top() << " ";
      // pq.pop();
   }
   cout << endl;
}

// 1-8 3-8 7-8 5-8 9-8 10-5 6-5 4-5 2-5 8-5

#endif
