// Purpose.  auto_ptr with exception lab
//
// Discussion.  An object's destructor will be called only if its constructor
// successfully ran to completion.  Below, the first two iterations worked fine,
// In the third iteration, the first inner object was successfully constructed,
// but the second inner object threw an exception.  That means that the outer
// object's constructor did NOT successfully complete.  As a result, the outer
// object's destructor should not be called, which means the first inner object
// is leaked!!  Wrapping dumb pointers in auto_ptrs can make constructors
// "safe", even in the presence of exceptions.
//
// Assignment.
// o Replace the two Inner pointer attributes in Outer with auto_ptrs
// o These two auto_ptrs are objects on the stack and will need to be initial-
//   ized with a member initialization list
// o Do you still need the delete statements?

#include <iostream>
#include <string>
using namespace std;

class Outer {
   class Inner {
      char value;
   public:
      Inner( char val ) {
         if (val == 'f') throw "out of luck";
         value = val;
         cout << "ctor-" << value << "  ";
      }
      ~Inner() { cout << value << "-dtor" << "  "; }
      void report() { cout << value << ' '; }
   };

   int   value;
   Inner *one, *two;
public:
   Outer( int val ) {
      value = val;
      one = new Inner( (val-1) * 2 + 'a' );
      two = new Inner( (val-1) * 2 + 'b' );
      cout << "ctor-" << value << '\n';
   }
   ~Outer() {
      cout << value << "-dtor" << "  ";
      delete two;
      delete one;
      cout << '\n';
   }
   void report() {
      cout << value << " -- ";
      one->report();
      two->report();
      cout << '\n';
   }
};

void main( void ) {
   try {
      for (int i=1; i < 4; i++) {
         Outer object( i );
         object.report();
      }
   } catch (char* ex) {
      cout << "\nmain - EXCEPTION - " << ex << '\n';
}  }

// ctor-a  ctor-b  ctor-1
// 1 -- a b
// 1-dtor  b-dtor  a-dtor
// ctor-c  ctor-d  ctor-2
// 2 -- c d
// 2-dtor  d-dtor  c-dtor
// ctor-e
// main - EXCEPTION - out of luck

//////// desired output /////////
// ctor-e  e-dtor
// main - EXCEPTION - out of luck
