// Purpose.  auto_ptr ownership lab answer

#include <iostream>
#include <memory>
using namespace std;

class Base { public:
   virtual ~Base() { }
   virtual void setValue( int ) = 0;
   virtual int  getValue() const = 0;
};

class A : public Base {
   int value;
public:
   A( int val )    { value = val;  cout << "Actor-" << value << '\n'; }
   A( const A& o ) { cout << "A copy ctor - " << o.value << '\n'; }
   ~A()            { cout << "dtorA-" << value << '\n'; }
   A& operator=( const A& rhs ) { cout << "A op= " << rhs.value << '\n'; }
   friend ostream& operator<< ( ostream& os, const A& obj ) {
      return os << "A-" << obj.value;
   }
   /*virtual*/ void setValue( int val ) { value = val; }
   /*virtual*/ int  getValue() const { return value; }
};

auto_ptr<Base> makeObject( int val ) {
   auto_ptr<Base> obj( new A( val ) );
   return obj;
}

void useObject( const auto_ptr<Base>& obj ) { obj->setValue( obj->getValue() +1 ); }

void ownObject( auto_ptr<Base> obj ) {
   obj->setValue( obj->getValue() + 1 );
   cout << "ownObject: value is " << obj->getValue() << '\n';
}

void main( void ) {
   auto_ptr<Base> obj( makeObject( 0 ) );
   for (int i=0; i < 4; i++) {
      useObject( obj );
      cout << "main: value is " << obj->getValue() << '\n';
   }
   ownObject( obj );
   cout << "main: done\n";
}

// Actor-0
// main: value is 1
// main: value is 2
// main: value is 3
// main: value is 4
// ownObject: value is 5
// dtorA-5
// main: done
