/* Generated by Together */

#include "MemoryManager.h"
#include "BitMap.h"
void * MemoryManager::allocate(unsigned long nbytes){
    unsigned long run_bits;
	long index;
	struct BiNode node;

    PRINT("MemoryManager::allocate(): request %lu bytes\n",nbytes);

    //translate nbytes into # of bits in BitMap

    run_bits = (nbytes/BYTES_PER_BITMAP_BIT)+1;

    PRINT("MemoryManager::allocate(): run_bits=%lu\n",run_bits);

    //look for # of free bits in BitMap

    index = ((*bmap).getBitRun(run_bits));

    PRINT("MemoryManager::allocate(): found run of %lu bits ",run_bits);
	PRINT("at %lu\n",index);

    if(index==-1){ return(NULL); }
	
    //reserved bits in BitMap

    (*bmap).setBits(0,run_bits,index);

    node.value = (unsigned long)(&mem[index*16]);
	node.index = index;
	node.nreserved = run_bits;

    bst.insertNode(&(bst.root_ptr),&node);

    //return memory represented by BitMap bits

    PRINT("MemoryManager::allocate(): address=%lu\n",&mem[index*16]); 

    return((void*)&mem[index*16]);
    }

void MemoryManager::release(void * addr){
    struct BiNode *ptr;

//    ptr = bst.findNode(bst.root_ptr,(unsigned long)addr);
    ptr = bst.findNode(bst.getRoot(),(unsigned long)addr);
	if(ptr!=NULL)
    {
        PRINT("MemoryManager::release(): address=%lu\n",(unsigned long)addr);
	    (*bmap).setBits(1,(*ptr).nreserved,(*ptr).index);
	    bst.deleteNode(&(bst.root_ptr),(unsigned long)addr);
	}
    return;
}

void MemoryManager::printState(){
    printf("-----------------------------------------------\n");
    (*bmap).printMap();
    printf("-----------------------------------------------\n");
//    bst.printTree(bst.root_ptr,0);
    bst.printTree(bst.getRoot(),0);
    printf("-----------------------------------------------\n");
    return;
}

MemoryManager::~MemoryManager(){
    //release resources for objects: bmap, bst, mem[]

    delete(bmap);

    bst.deleteAll(&(bst.root_ptr));

    if(HeapFree(handle,HEAP_NO_SERIALIZE,mem)==0)
    {
        printf("MemoryManager::~MemoryManager(): HeapFree() failed\n");
	    return;
	}

    //for portability, you could use:
	//free(mem);
    
    printf("MemoryManager::~MemoryManager():");
    printf("free() mem[%lu]\n",memLength);

    return;

}

MemoryManager::MemoryManager(unsigned long totalbytes){
    //init 3 dynamic objects: bmap, bst, mem[]

    bmap = new BitMap((totalbytes/BYTES_PER_BITMAP_BYTE)+1);

    bst.root_ptr=NULL;

    memLength = (*bmap).getByteSize()*BYTES_PER_BITMAP_BYTE;

    handle = GetProcessHeap();
	if(handle==NULL)
    {
        printf("MemoryManager::MemoryManager(): invalid handle\n");
	    return;
	}

    mem = (unsigned char*)HeapAlloc(handle,HEAP_ZERO_MEMORY,memLength);
	
    //for portability, you could use:
	//mem = (unsigned char*)malloc(memLength);

    if(mem==NULL)
    {
        printf("MemoryManager::MemoryManager():");
        printf("could not alloc memory\n");
	    exit(1);
	}
    printf("MemoryManager::MemoryManager():"); 
	printf("mallloc() mem[%lu]\n",memLength);

    return;

}
