#ifndef __MEMMGR_H__
#define __MEMMGR_H__
#include "common.h"
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	name: SequentialFitMemoryManager
	use:  uses manual memory manager to allocate/free memory
	      utilizes a linked-list, first-fit, scheme

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
	
#ifdef  DEBUG_SF_MEM_MGR
#define MSG0(arg);			printf(arg);
#define MSG1(arg1,arg2);	printf(arg1,arg2);
#else
#define MSG0(arg);		
#define MSG1(arg1,arg2);	
#endif

#define U1 unsigned char
#define U4 unsigned long

/*
	list element format  
	                    |0  3||4  7||  8  ||9 12||13 .. n|
						[PREV][NEXT][STATE][SIZE][payload]
	                       U4    U4    U1    U4     ?

	byte allocated/freed is address of first byte of payload
	header = 13 bytes

	byte[0] is occupied by header data, so is always used, thus
			first link has prev=0 ( 0 indicates not used )
			last link has next=0
*/

#define PREV(i)		(*((U4*)(&ram[i-13])))
#define NEXT(i)		(*((U4*)(&ram[i-9])))
#define STATE(i)	(*((U1*)(&ram[i-5])))  /*FREE,OCCUPIED*/
#define SIZE(i)		(*((U4*)(&ram[i-4])))

#define FREE				0
#define OCCUPIED			1
extern char *stateStr[3];

#define START		13	/*address of first payload*/
#define SZ_HEADER	13

class SequentialFitMemoryManager
{
	private:
	HANDLE handle;
	U1  *ram;	/*memory storage*/
	U4  size;	

	void split(U4 addr,U4 nbytes);
	void merge(U4 prev,U4 current,U4 next);

	public:

	SequentialFitMemoryManager(U4 nbytes);
	~SequentialFitMemoryManager();
	
	void*allocate(U4 nbytes);
	void release(void* addr);
	void printState();
};


#endif //__MEMMGR_H__