#include "memmgr.h"

char *stateStr[3]={"FREE","OCCUPIED"};

SegregatedMemoryManager::SegregatedMemoryManager(U4 nbytes)
{
	handle = GetProcessHeap();
	if(handle==NULL)
	{
		printf("SegregatedMemoryManager::");
		printf("SegregatedMemoryManager():");
		printf("invalid handle\n");
		exit(1);
	}

	ram = (U1*)HeapAlloc(handle,HEAP_ZERO_MEMORY,nbytes);
	
	//for portability, you could use:
	//ram = (unsigned char*)malloc(nbytes);


	size = nbytes;

	if(size<SZ_ROW)
	{
		printf("SegregatedMemoryManager::");
		printf("SegregatedMemoryManager():");
		printf("not enough memory fed to constructor\n");
		exit(1);	
	}

	nrows = size/SZ_ROW;

	initColumn(START16);
	initColumn(START32);
	initColumn(START64);
	initColumn(START128);
	initColumn(START256);
	initColumn(START512);
	initColumn(START1024);
	initColumn(START4096);

	MSG0("SegregatedMemoryManager::");
	MSG1("SegregatedMemoryManager(%lu)\n",nbytes);

	return;

}/*end constructor-------------------------------------------------*/

void SegregatedMemoryManager::initColumn(U4 index)
{
	U4 i;
	for(i=0;i<nrows;i++)
	{
		STATE(index)=FREE;
		index = index + SZ_ROW;
	}
	return;

}/*end initColumn--------------------------------------------------*/

SegregatedMemoryManager::~SegregatedMemoryManager()
{
	if(HeapFree(handle,HEAP_NO_SERIALIZE,ram)==0)
	{
		printf("SegregatedMemoryManager::");
		printf("~SegregatedMemoryManager():");
		printf("could not free heap storage\n");
		return;
	}

	//for portability, you could use:
	//free(ram);

	MSG0("SegregatedMemoryManager::");
	MSG0("~SegregatedFitMemoryManager()");
	MSG1("free ram[%lu]\n",size);
	return;

}/*end destructor--------------------------------------------------*/

/*
U4 nbytes	-	number of bytes required
returns address of first byte of memory region allocated 
( or NULL if cannot allocate a large enough block )
*/

void* SegregatedMemoryManager::allocate(U4 nbytes)
{
	U4 index;

	MSG0("SegregatedMemoryManager::");
	MSG1("allocate(%lu)\n",nbytes);

	if(nbytes==0)
	{ 
		MSG0("SegregatedMemoryManager::");
		MSG0("allocate(): zero bytes requested\n");
		return(NULL);
	}

	if(nbytes<=16)
	{
		index = searchColumn(START16);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START32);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START64);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START128);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START256);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START512);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START1024);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START4096);
		if(index){ return((void*)&ram[index]); }
	}
	else if(nbytes<=32)
	{
		index = searchColumn(START32);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START64);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START128);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START256);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START512);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START1024);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START4096);
		if(index){ return((void*)&ram[index]); }
	}
	else if(nbytes<=64)
	{
		index = searchColumn(START64);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START128);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START256);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START512);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START1024);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START4096);
		if(index){ return((void*)&ram[index]); }
	}
	else if(nbytes<=128)
	{
		index = searchColumn(START128);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START256);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START512);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START1024);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START4096);
		if(index){ return((void*)&ram[index]); }
	}
	else if(nbytes<=256)
	{
		index = searchColumn(START256);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START512);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START1024);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START4096);
		if(index){ return((void*)&ram[index]); }
	}
	else if(nbytes<=512)
	{
		index = searchColumn(START512);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START1024);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START4096);
		if(index){ return((void*)&ram[index]); }
	}
	else if(nbytes<=1024)
	{
		index = searchColumn(START1024);
		if(index){ return((void*)&ram[index]); }
		index = searchColumn(START4096);
		if(index){ return((void*)&ram[index]); }
	}
	else if(nbytes<=4096)
	{
		index = searchColumn(START4096);
		if(index){ return((void*)&ram[index]); }
	}

	return(NULL);

}/*end allocation--------------------------------------------------*/

/*
search a given size range for a free element
return index 
or zero if no available memory
*/

U4 SegregatedMemoryManager::searchColumn(U4 index)
{
	U4 i;
	for(i=0;i<nrows;i++)
	{
		if(STATE(index)==FREE)
		{ 
			MSG0("SegregatedMemoryManager::");
			MSG1("searchColumn(): free at index %lu, ",index);
			MSG1("address=%p\n",&ram[index]);
			STATE(index)=OCCUPIED;
			return(index); 
		}
		index = index + SZ_ROW;
	}
	return(0);

}/*end searchColumn------------------------------------------------*/

void SegregatedMemoryManager::release(void *addr)
{
	U4 free;	//index into ram[]

	if(addr==NULL)
	{
		MSG0("SegregatedMemoryManager::");
		MSG0("release(): cannot release NULL pointer\n");
		return;
	}

	MSG0("SegregatedMemoryManager::");
	MSG1("release(%p)\n",addr);

	//perform sanity check to make sure address is kosher

	if( (addr>= (void*)&ram[size]) || (addr< (void*)&ram[0]) )
	{
		MSG0("SegregatedMemoryManager::");
		MSG0("release(): address out of bounds\n");
		return;
	}

	//translate void* addr to index in ram[]

	free = (U4)( ((U1*)addr) - &ram[0] );

	MSG0("SegregatedMemoryManager::");
	MSG1("address resolves to index %lu\n",free);

	//a header always occupies first 13 bytes of storage

	if(free==0)
	{
		MSG0("SegregatedMemoryManager::");
		MSG0("release(): address in first 1st byte\n");
		return;	
	}

	//yet more sanity checks

	if(STATE(free)!=OCCUPIED)
	{
		MSG0("SegregatedMemoryManager::");
		MSG0("release(): referencing invalid region\n");
		return;	
	}

	STATE(free)=FREE;

	return;

}/*end release-----------------------------------------------------*/

void SegregatedMemoryManager::printState()
{
	printf("[16 bytes]");
	printColumn(START16);

	printf("[32 bytes]");
	printColumn(START32);

	printf("[64 bytes]");
	printColumn(START64);

	printf("[128 bytes]");
	printColumn(START128);

	printf("[256 bytes]");
	printColumn(START256);

	printf("[512 bytes]");
	printColumn(START512);

	printf("[1024 bytes]");
	printColumn(START1024);

	printf("[4096 bytes]");
	printColumn(START4096);

	return;

}/*end printState--------------------------------------------------*/

void SegregatedMemoryManager::printColumn(U4 index)
{
	U4 i;
	for(i=0;i<nrows;i++)
	{
		if(STATE(index)==OCCUPIED)
		{ 
			printf("[%p] ",&ram[index]); 
		}
		index = index + SZ_ROW;
	}
	printf("\n");
	return;

}/*end printColumn-------------------------------------------------*/

